User's Guide
------------

Usually we have no time and desire to read lengthy docs, so this tutorial is compact 
and has commented code examples for all of the common use cases.
If you are interested in some internal Sphinxit mechanisms, feel free to explore tests and sources.

.. toctree::
   :maxdepth: 2

   installation
   preparation
   usage
   integration

.. toctree::
   :maxdepth: 2

   changelog


Development
-----------
You can help to improve Sphinxit by submitting `issues` and `pull requests`
on `Github project page <https://github.com/semirook/sphinxit>`_.


Contacts
--------

If you have some private proposals or questions,
find me in `Twitter <https://twitter.com/semirook>`_ or `Facebook <http://www.facebook.com/semirook>`_.
